/* Terraform - (C) 1997-2000 Robert Gasch (r.gasch@chello.nl)
 *  - http://212.187.12.197/RNG/terraform/
 *  - http://terraform.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef _FLEX_ARRAY_H
#define _FLEX_ARRAY_H 1

#include "glib.h"
#include "string.h"
#include "GlobalSanityCheck.h"


/// This should really be a template class. 
/**
 *  FlexArray: an array which automatically extends (using ever bigger memory 
 *	chunks) as new data requires and provides fast data access.
 */
class FlexArray
	{
	public:
				FlexArray (int initialSize=100, float growthFactor=2.0);
		virtual		~FlexArray ();
		void 		append (void *ptr);
		void 		*El (int i);
		void		*first ();
		void		*last ();
		void		trim ();
		int		remove (int loc, int n=1 );
		int		removeLast (int n=1);
		int		getSize () { return d_nElements; }
		void		*getData () { return p_array; }
		void		setMaxDelIter (int n) { d_maxDelIter = n; }

	protected:
		void		extendArray ();

		/// provided so the data elements can be deleted in a type-safe manner
		virtual void	clearArray () 		{;}

		void			**p_array;
		int			d_nElements,
					d_len,
					d_inc,
					d_maxDelIter;
		float			d_incFactor;

	// C++ sugar
	public:
		bool operator==(const FlexArray & rhs)
			{
			if (p_array && rhs.p_array)
				return (p_array==rhs.p_array);
			return (FALSE);
			}
	};


/**
 *  Constructor: initialize everything; data array is not allocated yet
 */
inline FlexArray::FlexArray(int initialSize, float growthFactor)
{
	p_array = NULL;
	d_inc = initialSize;
	d_incFactor = growthFactor;
	d_nElements = 0;
	d_len = 0;
	d_maxDelIter = initialSize;
}


/**
 *  Destructor: free the allocated memory. 
 */
inline FlexArray::~FlexArray()
{
	if (p_array)
		{
		clearArray ();
		delete [] p_array;	// user must implements this class
		}
}


/**
 *  El: return the requested array element; otherwise NULL
 */
inline void *FlexArray::El (int i)
{
	if (i > d_nElements-1)
		return NULL;
	else
		return p_array[i];
}


/**
 *  first: return the first element; otherwise NULL
 */
inline void *FlexArray::first ()
{
	if (!p_array)
		return NULL;
	else
		return p_array[0];
}


/**
 *  last: return the last element; otherwise NULL
 */
inline void *FlexArray::last ()
{
	if (!p_array)
		return NULL;
	else
		return p_array[d_nElements-1];
}


/**
 *  append: append the pointer to the data array.
 */
inline void FlexArray::append (void *ptr)
{
	// check if we have enough space
	if (d_nElements >= d_len)
		extendArray ();

	// insert;
	p_array[d_nElements++] = ptr;
}


/**
 *  remove: remove the specified (series of) element(s). Does not care 
 * 	about memory so you should delete the objects before removing them. 
 */
inline int FlexArray::remove (int loc, int n)
{
	int		i;

	if (d_nElements <= loc)
		return -1;

	if (loc+n > d_nElements)
		return -1;

	if ((loc == (d_nElements-n)))
		{
		for (i=0; i<n; i++)
			p_array[loc] = NULL;
		}
	else
		{
		// remove iterativeley
		if (n <= d_maxDelIter)
			{
			int	lim = d_nElements - loc - n;
			for (i=0; i<lim; i++)
			p_array[loc+i] = p_array[loc+n+i];
			}
		// remove by using memcpy
		else
			{
			void	**newArr = NULL;
			int	off = loc+n;

			newArr = new void*[d_len];
			memcpy (newArr, p_array, loc*sizeof(void*));
			memcpy (newArr+loc, p_array+off, (d_len-off)*sizeof(void*));
			if (p_array)
				delete [] p_array;
			p_array = newArr;
			}
		}

	d_nElements-=n;
	return 0;
}


/**
 *  removeLast: remove the last n elements. Does not care about memory
 * 	so you should delete the objects before removing them. 
 */
inline int FlexArray::removeLast (int n)
{
	if (!d_nElements)
		return -1;

	for (int i=d_nElements-n; i<d_nElements; i++)
		p_array[i] = NULL;

	d_nElements-=n;
	return 0;
}
		

/**
 *  trim: shrink the data array to fit the current number of objects it holds. 
 */
inline void FlexArray::trim ()
{
	void	**newArr = NULL;

	if (!p_array)
		return;

	if (d_nElements == d_len)
		return;

	newArr = new void*[d_nElements];
	memcpy (newArr, p_array, d_nElements*sizeof(void*));
	delete [] p_array;
	d_len = d_nElements;
	p_array = newArr;
}


/**
 *  extendArray: allocate a new/larger data array and copy the existing 
 * 	array into the new one. Then delete the original data array. 
 */
inline void FlexArray::extendArray ()
{
	void	**newArr = NULL;
	int	blen = d_nElements*sizeof(void*); // byte length of used area

	//printf ("Allocating %d array cells\n", d_inc);
	newArr = new void*[d_len+d_inc];
	memcpy (newArr, p_array, blen);
	memset (newArr+d_len, 0, d_inc*sizeof(void*));
	if (p_array)
		delete [] p_array;
	d_len += d_inc;
	d_inc = (int)(d_inc*d_incFactor);
	p_array = newArr;
}

#endif // _FLEX_ARRAY_H
